import { B as BaseStorageOptions, I as InitSQLiteOptions } from './index-DoihFDbg.js';

declare class OPFSAnyContextVFS {
    static create(name: any, module: any, options: any, root: any): Promise<OPFSAnyContextVFS>;
    constructor(name: any, module: any, options?: {});
    /** @type {Map<number, File>} */ mapIdToFile: Map<number, File>;
    lastError: null;
    log: null;
    /**
     * HACK: root directory handle, to match `isFsHandleVFS` check
    */
    releaser: null;
    /**
     * @param {string} path
     * @param {boolean|undefined} create
     * @returns {Promise<[FileSystemDirectoryHandle, string]>}
     */
    getPathComponents(path: string, create: boolean | undefined): Promise<[FileSystemDirectoryHandle, string]>;
    getFilename(fileId: any): string;
    /**
     * @param {string?} zName
     * @param {number} fileId
     * @param {number} flags
     * @param {DataView} pOutFlags
     * @returns {Promise<number>}
     */
    jOpen(zName: string | null, fileId: number, flags: number, pOutFlags: DataView): Promise<number>;
    /**
     * @param {string} zName
     * @param {number} syncDir
     * @returns {Promise<number>}
     */
    jDelete(zName: string, syncDir: number): Promise<number>;
    /**
     * @param {string} zName
     * @param {number} flags
     * @param {DataView} pResOut
     * @returns {Promise<number>}
     */
    jAccess(zName: string, flags: number, pResOut: DataView): Promise<number>;
    /**
     * @param {number} fileId
     * @returns {Promise<number>}
     */
    jClose(fileId: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {Uint8Array} pData
     * @param {number} iOffset
     * @returns {Promise<number>}
     */
    jRead(fileId: number, pData: Uint8Array, iOffset: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {Uint8Array} pData
     * @param {number} iOffset
     * @returns {Promise<number>}
     */
    jWrite(fileId: number, pData: Uint8Array, iOffset: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} iSize
     * @returns {Promise<number>}
     */
    jTruncate(fileId: number, iSize: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} flags
     * @returns {Promise<number>}
     */
    jSync(fileId: number, flags: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {DataView} pSize64
     * @returns {Promise<number>}
     */
    jFileSize(fileId: number, pSize64: DataView): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} lockType
     * @returns {Promise<number>}
     */
    jLock(fileId: number, lockType: number): Promise<number>;
    jGetLastError(zBuf: any): 0;
}
declare class File {
    constructor(pathname: any, flags: any);
    /** @type {string} */ pathname: string;
    /** @type {number} */ flags: number;
    /** @type {FileSystemFileHandle} */ fileHandle: FileSystemFileHandle;
    /** @type {Blob?} */ blob: Blob | null;
    /** @type {FileSystemWritableFileStream?} */ writable: FileSystemWritableFileStream | null;
}

/**
 * Store data through `FileSystemFileHandle`,
 * use modified `OPFSAnyContextVFS` with `wa-sqlite-async.wasm`,
 * allow to use device's local file or OPFS in main or worker thread,
 * but a little slower than `useOpfsStorage`
 *
 * @param path db file directory path
 * @param rootDirHandle root directory handle
 * @param options wasm file url
 * @example
 * ```ts
 * import { initSQLite, isOpfsSupported } from '@subframe7536/sqlite-wasm'
 * import { useFsHandleStorage } from '@subframe7536/sqlite-wasm/fs-handle'
 *
 * // optional url
 * const url = 'https://cdn.jsdelivr.net/npm/@subframe7536/sqlite-wasm@0.5.0/dist/wa-sqlite-async.wasm'
 *
 * // device's local file
 * const root = await window.showDirectoryPicker()
 * // OPFS
 * const root1 = await navigator.storage.getDirectory()
 *
 * const { run, changes, lastInsertRowId, close } = await initSQLite(
 *   useFsHandleStorage('test.db', root, url)
 * )
 * ```
 */
declare function useFsHandleStorage(path: string, rootDirHandle: FileSystemDirectoryHandle, options?: BaseStorageOptions): Promise<InitSQLiteOptions>;

export { OPFSAnyContextVFS, useFsHandleStorage };
