import { B as BaseStorageOptions, I as InitSQLiteOptions } from './index-DoihFDbg.cjs';

declare class IDBMirrorVFS {
    static create(name: any, module: any, options: any): Promise<IDBMirrorVFS>;
    constructor(name: any, module: any, options?: {});
    log: null;
    close(): void;
    isReady(): Promise<any>;
    /**
     * @param {string?} zName
     * @param {number} fileId
     * @param {number} flags
     * @param {DataView} pOutFlags
     * @returns {Promise<number>}
     */
    jOpen(zName: string | null, fileId: number, flags: number, pOutFlags: DataView): Promise<number>;
    /**
     * @param {string} zName
     * @param {number} syncDir
     * @returns {Promise<number>}
     */
    jDelete(zName: string, syncDir: number): Promise<number>;
    /**
     * @param {string} zName
     * @param {number} flags
     * @param {DataView} pResOut
     * @returns {Promise<number>}
     */
    jAccess(zName: string, flags: number, pResOut: DataView): Promise<number>;
    /**
     * @param {number} fileId
     * @returns {Promise<number>}
     */
    jClose(fileId: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {Uint8Array} pData
     * @param {number} iOffset
     * @returns {number}
     */
    jRead(fileId: number, pData: Uint8Array, iOffset: number): number;
    /**
     * @param {number} fileId
     * @param {Uint8Array} pData
     * @param {number} iOffset
     * @returns {number}
     */
    jWrite(fileId: number, pData: Uint8Array, iOffset: number): number;
    lastError: unknown;
    /**
     * @param {number} fileId
     * @param {number} iSize
     * @returns {number}
     */
    jTruncate(fileId: number, iSize: number): number;
    /**
     * @param {number} fileId
     * @param {DataView} pSize64
     * @returns {number|Promise<number>}
     */
    jFileSize(fileId: number, pSize64: DataView): number | Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} lockType
     * @returns {Promise<number>}
     */
    jLock(fileId: number, lockType: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} lockType
     * @returns {number}
     */
    jUnlock(fileId: number, lockType: number): number;
    /**
     * @param {number} fileId
     * @param {DataView} pResOut
     * @returns {Promise<number>}
     */
    jCheckReservedLock(fileId: number, pResOut: DataView): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} op
     * @param {DataView} pArg
     * @returns {Promise<number>}
     */
    jFileControl(fileId: number, op: number, pArg: DataView): Promise<number>;
    /**
     * @param {number} fileId
     * @returns {number|Promise<number>}
     */
    jDeviceCharacteristics(fileId: number): number | Promise<number>;
    /**
     * @param {Uint8Array} zBuf
     * @returns {number}
     */
    jGetLastError(zBuf: Uint8Array): number;
    #private;
}

/**
 * Store data in memory and sync to `IndexedDB`,
 * use `IDBMirrorVFS` with `wa-sqlite-async.wasm` (larger than sync version),
 * better performance compare to `useIdbStorage`.
 * @param fileName db file name
 * @param options options
 * @example
 * ```ts
 * import { initSQLite } from '@subframe7536/sqlite-wasm'
 * import { useIdbMemoryStorage } from '@subframe7536/sqlite-wasm/idb-memory'
 *
 * // optional url
 * const url = 'https://cdn.jsdelivr.net/npm/@subframe7536/sqlite-wasm@0.5.0/dist/wa-sqlite-async.wasm'
 *
 * const { run, changes, lastInsertRowId, close } = await initSQLite(
 *   useIdbMemoryStorage('test.db', { url })
 * )
 * ```
 */
declare function useIdbMemoryStorage(fileName: string, options?: BaseStorageOptions): Promise<InitSQLiteOptions>;

export { IDBMirrorVFS, useIdbMemoryStorage };
