import { b as IDBBatchAtomicVFSOptions, B as BaseStorageOptions, I as InitSQLiteOptions } from './index-DoihFDbg.js';

declare class IDBBatchAtomicVFS {
    static create(name: any, module: any, options: any): Promise<IDBBatchAtomicVFS>;
    constructor(name: any, module: any, options?: {});
    /** @type {Map<number, File>} */ mapIdToFile: Map<number, File>;
    lastError: null;
    log: null;
    close(): void;
    isReady(): Promise<void>;
    getFilename(fileId: any): string;
    /**
     * @param {string?} zName
     * @param {number} fileId
     * @param {number} flags
     * @param {DataView} pOutFlags
     * @returns {Promise<number>}
     */
    jOpen(zName: string | null, fileId: number, flags: number, pOutFlags: DataView): Promise<number>;
    /**
     * @param {string} zName
     * @param {number} syncDir
     * @returns {Promise<number>}
     */
    jDelete(zName: string, syncDir: number): Promise<number>;
    /**
     * @param {string} zName
     * @param {number} flags
     * @param {DataView} pResOut
     * @returns {Promise<number>}
     */
    jAccess(zName: string, flags: number, pResOut: DataView): Promise<number>;
    /**
     * @param {number} fileId
     * @returns {Promise<number>}
     */
    jClose(fileId: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {Uint8Array} pData
     * @param {number} iOffset
     * @returns {Promise<number>}
     */
    jRead(fileId: number, pData: Uint8Array, iOffset: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {Uint8Array} pData
     * @param {number} iOffset
     * @returns {number}
     */
    jWrite(fileId: number, pData: Uint8Array, iOffset: number): number;
    /**
     * @param {number} fileId
     * @param {number} iSize
     * @returns {number}
     */
    jTruncate(fileId: number, iSize: number): number;
    /**
     * @param {number} fileId
     * @param {number} flags
     * @returns {Promise<number>}
     */
    jSync(fileId: number, flags: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {DataView} pSize64
     * @returns {number}
     */
    jFileSize(fileId: number, pSize64: DataView): number;
    /**
     * @param {number} fileId
     * @param {number} lockType
     * @returns {Promise<number>}
     */
    jLock(fileId: number, lockType: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} lockType
     * @returns {Promise<number>}
     */
    jUnlock(fileId: number, lockType: number): Promise<number>;
    /**
     * @param {number} fileId
     * @param {number} op
     * @param {DataView} pArg
     * @returns {number|Promise<number>}
     */
    jFileControl(fileId: number, op: number, pArg: DataView): number | Promise<number>;
    /**
     * @param {number} pFile
     * @returns {number|Promise<number>}
     */
    jDeviceCharacteristics(pFile: number): number | Promise<number>;
    /**
     * @param {Uint8Array} zBuf
     * @returns {number|Promise<number>}
     */
    jGetLastError(zBuf: Uint8Array): number | Promise<number>;
    #private;
}
type Metadata = {
    name: string;
    fileSize: number;
    version: number;
    pendingVersion?: number | undefined;
};
/**
 * @typedef Metadata
 * @property {string} name
 * @property {number} fileSize
 * @property {number} version
 * @property {number} [pendingVersion]
 */
declare class File {
    constructor(path: any, flags: any, metadata: any);
    /** @type {string} */ path: string;
    /** @type {number} */ flags: number;
    /** @type {Metadata} */ metadata: Metadata;
    /** @type {number} */ fileSize: number;
    /** @type {boolean} */ needsMetadataSync: boolean;
    /** @type {Metadata} */ rollback: Metadata;
    /** @type {Set<number>} */ changedPages: Set<number>;
    /** @type {string} */ synchronous: string;
    /** @type {IDBTransactionOptions} */ txOptions: IDBTransactionOptions;
}

type IDBVFSOptions = IDBBatchAtomicVFSOptions;
/**
 * Store data in `IndexedDB`,
 * use `IDBBatchAtomicVFS` with `wa-sqlite-async.wasm`,
 * larger than sync version, better compatibility
 * @param fileName db file name
 * @param options options
 * @example
 * ```ts
 * import { initSQLite } from '@subframe7536/sqlite-wasm'
 * import { useIdbStorage } from '@subframe7536/sqlite-wasm/idb'
 *
 * // optional url
 * const url = 'https://cdn.jsdelivr.net/npm/@subframe7536/sqlite-wasm@0.5.0/wa-sqlite-async.wasm'
 * const url1 = 'https://cdn.jsdelivr.net/gh/subframe7536/sqlite-wasm@v0.5.0/wa-sqlite-fts5/wa-sqlite-async.wasm'
 *
 * const { run, changes, lastInsertRowId, close } = await initSQLite(
 *   useIdbStorage('test.db', { url })
 * )
 * ```
 */
declare function useIdbStorage(fileName: string, options?: IDBVFSOptions & BaseStorageOptions): Promise<InitSQLiteOptions>;

export { IDBBatchAtomicVFS, type IDBVFSOptions, useIdbStorage };
