import { WebLocksMixin } from './chunk-JKVIPQ3X.js';
import { wa_sqlite_async_default } from './chunk-5LYQPZP3.js';
import { FacadeVFS } from './chunk-BESRLAWD.js';
import { SQLITE_OPEN_CREATE, SQLITE_OK, SQLITE_CANTOPEN, SQLITE_IOERR_DELETE, SQLITE_IOERR_ACCESS, SQLITE_OPEN_DELETEONCLOSE, SQLITE_IOERR_CLOSE, SQLITE_IOERR_SHORT_READ, SQLITE_IOERR_READ, SQLITE_OPEN_MAIN_DB, SQLITE_OPEN_TEMP_DB, SQLITE_IOERR_WRITE, SQLITE_IOERR_TRUNCATE, SQLITE_IOERR_FSYNC, SQLITE_IOERR_FSTAT, SQLITE_LOCK_SHARED, SQLITE_LOCK_NONE, SQLITE_FCNTL_ROLLBACK_ATOMIC_WRITE, SQLITE_FCNTL_COMMIT_ATOMIC_WRITE, SQLITE_FCNTL_BEGIN_ATOMIC_WRITE, SQLITE_FCNTL_SYNC, SQLITE_FCNTL_PRAGMA, SQLITE_ERROR, SQLITE_IOERR, SQLITE_IOCAP_BATCH_ATOMIC, SQLITE_IOCAP_UNDELETABLE_WHEN_OPEN } from './chunk-6A5XENKU.js';

// node_modules/.pnpm/wa-sqlite@https+++codeload.github.com+rhashimoto+wa-sqlite+tar.gz+03c00ed67cd934bd664ae310234bba317928f851/node_modules/wa-sqlite/src/examples/IDBBatchAtomicVFS.js
var RETRYABLE_ERRORS = /* @__PURE__ */ new Set([
  "TransactionInactiveError",
  "InvalidStateError"
]);
var File = class {
  /** @type {string} */
  path;
  /** @type {number} */
  flags;
  /** @type {Metadata} */
  metadata;
  /** @type {number} */
  fileSize = 0;
  /** @type {boolean} */
  needsMetadataSync = false;
  /** @type {Metadata} */
  rollback = null;
  /** @type {Set<number>} */
  changedPages = /* @__PURE__ */ new Set();
  /** @type {string} */
  synchronous = "full";
  /** @type {IDBTransactionOptions} */
  txOptions = { durability: "strict" };
  constructor(path, flags, metadata) {
    this.path = path;
    this.flags = flags;
    this.metadata = metadata;
  }
};
var IDBBatchAtomicVFS = class _IDBBatchAtomicVFS extends WebLocksMixin(FacadeVFS) {
  /** @type {Map<number, File>} */
  mapIdToFile = /* @__PURE__ */ new Map();
  lastError = null;
  log = null;
  // console.log
  /** @type {Promise} */
  #isReady;
  /** @type {IDBContext} */
  #idb;
  static async create(name, module, options) {
    const vfs = new _IDBBatchAtomicVFS(name, module, options);
    await vfs.isReady();
    return vfs;
  }
  constructor(name, module, options = {}) {
    super(name, module, options);
    this.#isReady = this.#initialize(options.idbName ?? name);
  }
  async #initialize(name) {
    this.#idb = await IDBContext.create(name);
  }
  close() {
    this.#idb.close();
  }
  async isReady() {
    await super.isReady();
    await this.#isReady;
  }
  getFilename(fileId) {
    const pathname = this.mapIdToFile.get(fileId).path;
    return `IDB(${this.name}):${pathname}`;
  }
  /**
   * @param {string?} zName 
   * @param {number} fileId 
   * @param {number} flags 
   * @param {DataView} pOutFlags 
   * @returns {Promise<number>}
   */
  async jOpen(zName, fileId, flags, pOutFlags) {
    try {
      const url = new URL(zName || Math.random().toString(36).slice(2), "file://");
      const path = url.pathname;
      let meta = await this.#idb.q(({ metadata }) => metadata.get(path));
      if (!meta && flags & SQLITE_OPEN_CREATE) {
        meta = {
          name: path,
          fileSize: 0,
          version: 0
        };
        await this.#idb.q(({ metadata }) => metadata.put(meta), "rw");
      }
      if (!meta) {
        throw new Error(`File ${path} not found`);
      }
      const file = new File(path, flags, meta);
      this.mapIdToFile.set(fileId, file);
      pOutFlags.setInt32(0, flags, true);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_CANTOPEN;
    }
  }
  /**
   * @param {string} zName 
   * @param {number} syncDir 
   * @returns {Promise<number>}
   */
  async jDelete(zName, syncDir) {
    try {
      const url = new URL(zName, "file://");
      const path = url.pathname;
      this.#idb.q(({ metadata, blocks }) => {
        const range = IDBKeyRange.bound([path, -Infinity], [path, Infinity]);
        blocks.delete(range);
        metadata.delete(path);
      }, "rw");
      if (syncDir) {
        await this.#idb.sync(false);
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_DELETE;
    }
  }
  /**
   * @param {string} zName 
   * @param {number} flags 
   * @param {DataView} pResOut 
   * @returns {Promise<number>}
   */
  async jAccess(zName, flags, pResOut) {
    try {
      const url = new URL(zName, "file://");
      const path = url.pathname;
      const meta = await this.#idb.q(({ metadata }) => metadata.get(path));
      pResOut.setInt32(0, meta ? 1 : 0, true);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_ACCESS;
    }
  }
  /**
   * @param {number} fileId 
   * @returns {Promise<number>}
   */
  async jClose(fileId) {
    try {
      const file = this.mapIdToFile.get(fileId);
      this.mapIdToFile.delete(fileId);
      if (file.flags & SQLITE_OPEN_DELETEONCLOSE) {
        await this.#idb.q(({ metadata, blocks }) => {
          metadata.delete(file.path);
          blocks.delete(IDBKeyRange.bound([file.path, 0], [file.path, Infinity]));
        }, "rw");
      }
      if (file.needsMetadataSync) {
        this.#idb.q(({ metadata }) => metadata.put(file.metadata), "rw");
      }
      await this.#idb.sync(file.synchronous === "full");
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_CLOSE;
    }
  }
  /**
   * @param {number} fileId 
   * @param {Uint8Array} pData 
   * @param {number} iOffset
   * @returns {Promise<number>}
   */
  async jRead(fileId, pData, iOffset) {
    try {
      const file = this.mapIdToFile.get(fileId);
      let pDataOffset = 0;
      while (pDataOffset < pData.byteLength) {
        const fileOffset = iOffset + pDataOffset;
        const block = await this.#idb.q(({ blocks }) => {
          const range = IDBKeyRange.bound([file.path, -fileOffset], [file.path, Infinity]);
          return blocks.get(range);
        });
        if (!block || block.data.byteLength - block.offset <= fileOffset) {
          pData.fill(0, pDataOffset);
          return SQLITE_IOERR_SHORT_READ;
        }
        const dst = pData.subarray(pDataOffset);
        const srcOffset = fileOffset + block.offset;
        const nBytesToCopy = Math.min(
          Math.max(block.data.byteLength - srcOffset, 0),
          dst.byteLength
        );
        dst.set(block.data.subarray(srcOffset, srcOffset + nBytesToCopy));
        pDataOffset += nBytesToCopy;
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_READ;
    }
  }
  /**
   * @param {number} fileId 
   * @param {Uint8Array} pData 
   * @param {number} iOffset
   * @returns {number}
   */
  jWrite(fileId, pData, iOffset) {
    try {
      const file = this.mapIdToFile.get(fileId);
      if (file.flags & SQLITE_OPEN_MAIN_DB) {
        if (!file.rollback) {
          const pending = Object.assign(
            { pendingVersion: file.metadata.version - 1 },
            file.metadata
          );
          this.#idb.q(({ metadata }) => metadata.put(pending), "rw", file.txOptions);
          file.rollback = Object.assign({}, file.metadata);
          file.metadata.version--;
        }
      }
      if (file.flags & SQLITE_OPEN_MAIN_DB) {
        file.changedPages.add(iOffset);
      }
      const data = pData.slice();
      const version = file.metadata.version;
      const isOverwrite = iOffset < file.metadata.fileSize;
      if (!isOverwrite || file.flags & SQLITE_OPEN_MAIN_DB || file.flags & SQLITE_OPEN_TEMP_DB) {
        const block = {
          path: file.path,
          offset: -iOffset,
          version,
          data: pData.slice()
        };
        this.#idb.q(({ blocks }) => {
          blocks.put(block);
          file.changedPages.add(iOffset);
        }, "rw", file.txOptions);
      } else {
        this.#idb.q(async ({ blocks }) => {
          const range = IDBKeyRange.bound(
            [file.path, -iOffset],
            [file.path, Infinity]
          );
          const block = await blocks.get(range);
          block.data.subarray(iOffset + block.offset).set(data);
          blocks.put(block);
        }, "rw", file.txOptions);
      }
      if (file.metadata.fileSize < iOffset + pData.length) {
        file.metadata.fileSize = iOffset + pData.length;
        file.needsMetadataSync = true;
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_WRITE;
    }
  }
  /**
   * @param {number} fileId 
   * @param {number} iSize 
   * @returns {number}
   */
  jTruncate(fileId, iSize) {
    try {
      const file = this.mapIdToFile.get(fileId);
      if (iSize < file.metadata.fileSize) {
        this.#idb.q(({ blocks }) => {
          const range = IDBKeyRange.bound(
            [file.path, -Infinity],
            [file.path, -iSize, Infinity]
          );
          blocks.delete(range);
        }, "rw", file.txOptions);
        file.metadata.fileSize = iSize;
        file.needsMetadataSync = true;
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_TRUNCATE;
    }
  }
  /**
   * @param {number} fileId 
   * @param {number} flags 
   * @returns {Promise<number>}
   */
  async jSync(fileId, flags) {
    try {
      const file = this.mapIdToFile.get(fileId);
      if (file.needsMetadataSync) {
        this.#idb.q(({ metadata }) => metadata.put(file.metadata), "rw", file.txOptions);
        file.needsMetadataSync = false;
      }
      if (file.flags & SQLITE_OPEN_MAIN_DB) {
        if (file.synchronous === "full") {
          await this.#idb.sync(true);
        }
      } else {
        await this.#idb.sync(file.synchronous === "full");
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_FSYNC;
    }
  }
  /**
   * @param {number} fileId 
   * @param {DataView} pSize64 
   * @returns {number}
   */
  jFileSize(fileId, pSize64) {
    try {
      const file = this.mapIdToFile.get(fileId);
      pSize64.setBigInt64(0, BigInt(file.metadata.fileSize), true);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_FSTAT;
    }
  }
  /**
   * @param {number} fileId 
   * @param {number} lockType 
   * @returns {Promise<number>}
   */
  async jLock(fileId, lockType) {
    const file = this.mapIdToFile.get(fileId);
    const result = await super.jLock(fileId, lockType);
    if (lockType === SQLITE_LOCK_SHARED) {
      file.metadata = await this.#idb.q(async ({ metadata, blocks }) => {
        const m = await metadata.get(file.path);
        if (m.pendingVersion) {
          console.warn(`removing failed transaction ${m.pendingVersion}`);
          await new Promise((resolve, reject) => {
            const range = IDBKeyRange.bound([m.name, -Infinity], [m.name, Infinity]);
            const request = blocks.openCursor(range);
            request.onsuccess = () => {
              const cursor = request.result;
              if (cursor) {
                const block = cursor.value;
                if (block.version < m.version) {
                  cursor.delete();
                }
                cursor.continue();
              } else {
                resolve();
              }
            };
            request.onerror = () => reject(request.error);
          });
          delete m.pendingVersion;
          metadata.put(m);
        }
        return m;
      }, "rw", file.txOptions);
    }
    return result;
  }
  /**
   * @param {number} fileId 
   * @param {number} lockType 
   * @returns {Promise<number>}
   */
  async jUnlock(fileId, lockType) {
    if (lockType === SQLITE_LOCK_NONE) {
      const file = this.mapIdToFile.get(fileId);
      await this.#idb.sync(file.synchronous === "full");
    }
    return super.jUnlock(fileId, lockType);
  }
  /**
   * @param {number} fileId
   * @param {number} op
   * @param {DataView} pArg
   * @returns {number|Promise<number>}
   */
  jFileControl(fileId, op, pArg) {
    try {
      const file = this.mapIdToFile.get(fileId);
      switch (op) {
        case SQLITE_FCNTL_PRAGMA:
          const key = extractString(pArg, 4);
          const value = extractString(pArg, 8);
          this.log?.("xFileControl", file.path, "PRAGMA", key, value);
          const setPragmaResponse = (response) => {
            const encoded = new TextEncoder().encode(response);
            const out = this._module._sqlite3_malloc(encoded.byteLength);
            const outArray = this._module.HEAPU8.subarray(out, out + encoded.byteLength);
            outArray.set(encoded);
            pArg.setUint32(0, out, true);
            return SQLITE_ERROR;
          };
          switch (key.toLowerCase()) {
            case "page_size":
              if (file.flags & SQLITE_OPEN_MAIN_DB) {
                if (value && file.metadata.fileSize) {
                  return SQLITE_ERROR;
                }
              }
              break;
            case "synchronous":
              if (value) {
                switch (value.toLowerCase()) {
                  case "0":
                  case "off":
                    file.synchronous = "off";
                    file.txOptions = { durability: "relaxed" };
                    break;
                  case "1":
                  case "normal":
                    file.synchronous = "normal";
                    file.txOptions = { durability: "relaxed" };
                    break;
                  case "2":
                  case "3":
                  case "full":
                  case "extra":
                    file.synchronous = "full";
                    file.txOptions = { durability: "strict" };
                    break;
                }
              }
              break;
            case "write_hint":
              return super.jFileControl(fileId, WebLocksMixin.WRITE_HINT_OP_CODE, null);
          }
          break;
        case SQLITE_FCNTL_SYNC:
          this.log?.("xFileControl", file.path, "SYNC");
          if (file.rollback) {
            const commitMetadata = Object.assign({}, file.metadata);
            const prevFileSize = file.rollback.fileSize;
            this.#idb.q(({ metadata, blocks }) => {
              metadata.put(commitMetadata);
              for (const offset of file.changedPages) {
                if (offset < prevFileSize) {
                  const range = IDBKeyRange.bound(
                    [file.path, -offset, commitMetadata.version],
                    [file.path, -offset, Infinity],
                    true
                  );
                  blocks.delete(range);
                }
              }
              file.changedPages.clear();
            }, "rw", file.txOptions);
            file.needsMetadataSync = false;
            file.rollback = null;
          }
          break;
        case SQLITE_FCNTL_BEGIN_ATOMIC_WRITE:
          this.log?.("xFileControl", file.path, "BEGIN_ATOMIC_WRITE");
          return SQLITE_OK;
        case SQLITE_FCNTL_COMMIT_ATOMIC_WRITE:
          this.log?.("xFileControl", file.path, "COMMIT_ATOMIC_WRITE");
          return SQLITE_OK;
        case SQLITE_FCNTL_ROLLBACK_ATOMIC_WRITE:
          this.log?.("xFileControl", file.path, "ROLLBACK_ATOMIC_WRITE");
          file.metadata = file.rollback;
          const rollbackMetadata = Object.assign({}, file.metadata);
          this.#idb.q(({ metadata, blocks }) => {
            metadata.put(rollbackMetadata);
            for (const offset of file.changedPages) {
              blocks.delete([file.path, -offset, rollbackMetadata.version - 1]);
            }
            file.changedPages.clear();
          }, "rw", file.txOptions);
          file.needsMetadataSync = false;
          file.rollback = null;
          return SQLITE_OK;
      }
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR;
    }
    return super.jFileControl(fileId, op, pArg);
  }
  /**
   * @param {number} pFile
   * @returns {number|Promise<number>}
   */
  jDeviceCharacteristics(pFile) {
    return 0 | SQLITE_IOCAP_BATCH_ATOMIC | SQLITE_IOCAP_UNDELETABLE_WHEN_OPEN;
  }
  /**
   * @param {Uint8Array} zBuf 
   * @returns {number|Promise<number>}
   */
  jGetLastError(zBuf) {
    if (this.lastError) {
      console.error(this.lastError);
      const outputArray = zBuf.subarray(0, zBuf.byteLength - 1);
      const { written } = new TextEncoder().encodeInto(this.lastError.message, outputArray);
      zBuf[written] = 0;
    }
    return SQLITE_OK;
  }
};
function extractString(dataView, offset) {
  const p = dataView.getUint32(offset, true);
  if (p) {
    const chars = new Uint8Array(dataView.buffer, p);
    return new TextDecoder().decode(chars.subarray(0, chars.indexOf(0)));
  }
  return null;
}
var IDBContext = class _IDBContext {
  /** @type {IDBDatabase} */
  #database;
  /** @type {Promise} */
  #chain = null;
  /** @type {Promise<any>} */
  #txComplete = Promise.resolve();
  /** @type {IDBRequest?} */
  #request = null;
  /** @type {WeakSet<IDBTransaction>} */
  #txPending = /* @__PURE__ */ new WeakSet();
  log = null;
  static async create(name) {
    const database = await new Promise((resolve, reject) => {
      const request = indexedDB.open(name, 6);
      request.onupgradeneeded = async (event) => {
        const db = request.result;
        if (event.oldVersion) {
          console.log(`Upgrading IndexedDB from version ${event.oldVersion}`);
        }
        switch (event.oldVersion) {
          case 0:
            db.createObjectStore("blocks", { keyPath: ["path", "offset", "version"] }).createIndex("version", ["path", "version"]);
          // fall through intentionally
          case 5:
            const tx = request.transaction;
            const blocks = tx.objectStore("blocks");
            blocks.deleteIndex("version");
            const metadata = db.createObjectStore("metadata", { keyPath: "name" });
            await new Promise((resolve2, reject2) => {
              let lastBlock = {};
              const request2 = tx.objectStore("blocks").openCursor();
              request2.onsuccess = () => {
                const cursor = request2.result;
                if (cursor) {
                  const block = cursor.value;
                  if (typeof block.offset !== "number" || block.path === lastBlock.path && block.offset === lastBlock.offset) {
                    cursor.delete();
                  } else if (block.offset === 0) {
                    metadata.put({
                      name: block.path,
                      fileSize: block.fileSize,
                      version: block.version
                    });
                    delete block.fileSize;
                    cursor.update(block);
                  }
                  lastBlock = block;
                  cursor.continue();
                } else {
                  resolve2();
                }
              };
              request2.onerror = () => reject2(request2.error);
            });
            break;
        }
      };
      request.onsuccess = () => resolve(request.result);
      request.onerror = () => reject(request.error);
    });
    return new _IDBContext(database);
  }
  constructor(database) {
    this.#database = database;
  }
  close() {
    this.#database.close();
  }
  /**
   * @param {(stores: Object.<string, IDBObjectStore>) => any} f 
   * @param {'ro'|'rw'} mode 
   * @returns {Promise<any>}
   */
  q(f, mode = "ro", options = {}) {
    const txMode = mode === "ro" ? "readonly" : "readwrite";
    const txOptions = Object.assign({
      /** @type {IDBTransactionDurability} */
      durability: "default"
    }, options);
    this.#chain = (this.#chain || Promise.resolve()).then(() => this.#q(f, txMode, txOptions));
    return this.#chain;
  }
  /**
   * @param {(stores: Object.<string, IDBObjectStore>) => any} f 
   * @param {IDBTransactionMode} mode 
   * @param {IDBTransactionOptions} options
   * @returns {Promise<any>}
   */
  async #q(f, mode, options) {
    let tx;
    if (this.#request && this.#txPending.has(this.#request.transaction) && this.#request.transaction.mode >= mode && this.#request.transaction.durability === options.durability) {
      tx = this.#request.transaction;
      if (this.#request.readyState === "pending") {
        await new Promise((resolve) => {
          this.#request.addEventListener("success", resolve, { once: true });
          this.#request.addEventListener("error", resolve, { once: true });
        });
      }
    }
    for (let i = 0; i < 2; ++i) {
      if (!tx) {
        await this.#txComplete;
        tx = this.#database.transaction(this.#database.objectStoreNames, mode, options);
        this.log?.("IDBTransaction open", mode);
        this.#txPending.add(tx);
        this.#txComplete = new Promise((resolve, reject) => {
          tx.addEventListener("complete", () => {
            this.log?.("IDBTransaction complete");
            this.#txPending.delete(tx);
            resolve();
          });
          tx.addEventListener("abort", () => {
            this.#txPending.delete(tx);
            reject(new Error("transaction aborted"));
          });
        });
      }
      try {
        const objectStores = [...tx.objectStoreNames].map((name) => {
          return [name, this.proxyStoreOrIndex(tx.objectStore(name))];
        });
        return await f(Object.fromEntries(objectStores));
      } catch (e) {
        if (!i && RETRYABLE_ERRORS.has(e.name)) {
          this.log?.(`${e.name}, retrying`);
          tx = null;
          continue;
        }
        throw e;
      }
    }
  }
  /**
   * Object store methods that return an IDBRequest, except for cursor
   * creation, are wrapped to return a Promise. In addition, the
   * request is used internally for chaining.
   * @param {IDBObjectStore} objectStore 
   * @returns 
   */
  proxyStoreOrIndex(objectStore) {
    return new Proxy(objectStore, {
      get: (target, property, receiver) => {
        const result = Reflect.get(target, property, receiver);
        if (typeof result === "function") {
          return (...args) => {
            const maybeRequest = Reflect.apply(result, target, args);
            if (maybeRequest instanceof IDBRequest && !property.endsWith("Cursor")) {
              this.#request = maybeRequest;
              maybeRequest.addEventListener("error", () => {
                console.error(maybeRequest.error);
                maybeRequest.transaction.abort();
              }, { once: true });
              return wrap(maybeRequest);
            }
            return maybeRequest;
          };
        }
        return result;
      }
    });
  }
  /**
   * @param {boolean} durable 
   */
  async sync(durable) {
    if (this.#chain) {
      await this.#chain;
      if (durable) {
        await this.#txComplete;
      }
      this.reset();
    }
  }
  reset() {
    this.#chain = null;
    this.#txComplete = Promise.resolve();
    this.#request = null;
  }
};
function wrap(request) {
  return new Promise((resolve, reject) => {
    request.onsuccess = () => resolve(request.result);
    request.onerror = () => reject(request.error);
  });
}

// src/vfs/idb.ts
async function useIdbStorage(fileName, options = {}) {
  const {
    url,
    lockPolicy = "shared+hint",
    lockTimeout = Infinity,
    ...rest
  } = options;
  const sqliteModule = await wa_sqlite_async_default(
    url ? { locateFile: () => url } : void 0
  );
  const idbName = fileName.endsWith(".db") ? fileName : `${fileName}.db`;
  const vfsOptions = { idbName, lockPolicy, lockTimeout };
  return {
    path: idbName,
    sqliteModule,
    vfsFn: IDBBatchAtomicVFS.create,
    vfsOptions,
    ...rest
  };
}

export { IDBBatchAtomicVFS, useIdbStorage };
