import { wa_sqlite_default } from './chunk-3QGEWUC2.js';
import { FacadeVFS } from './chunk-BESRLAWD.js';
import { SQLITE_OPEN_MAIN_DB, SQLITE_BUSY, SQLITE_OPEN_CREATE, SQLITE_CANTOPEN, SQLITE_OK, SQLITE_IOERR_DELETE, SQLITE_IOERR_ACCESS, SQLITE_OPEN_DELETEONCLOSE, SQLITE_IOERR_CLOSE, SQLITE_IOERR_SHORT_READ, SQLITE_IOERR_READ, SQLITE_IOERR_WRITE, SQLITE_IOERR_TRUNCATE, SQLITE_IOERR_FSYNC, SQLITE_IOERR_FSTAT, SQLITE_LOCK_NONE, SQLITE_FCNTL_PRAGMA, SQLITE_IOERR, SQLITE_NOTFOUND } from './chunk-6A5XENKU.js';

// node_modules/.pnpm/wa-sqlite@https+++codeload.github.com+rhashimoto+wa-sqlite+tar.gz+03c00ed67cd934bd664ae310234bba317928f851/node_modules/wa-sqlite/src/examples/OPFSCoopSyncVFS.js
var DEFAULT_TEMPORARY_FILES = 10;
var LOCK_NOTIFY_INTERVAL = 1e3;
var DB_RELATED_FILE_SUFFIXES = ["", "-journal", "-wal"];
var finalizationRegistry = new FinalizationRegistry((releaser) => releaser());
var File = class {
  /** @type {string} */
  path;
  /** @type {number} */
  flags;
  /** @type {FileSystemSyncAccessHandle} */
  accessHandle;
  /** @type {PersistentFile?} */
  persistentFile;
  constructor(path, flags) {
    this.path = path;
    this.flags = flags;
  }
};
var PersistentFile = class {
  /** @type {FileSystemFileHandle} */
  fileHandle;
  /** @type {FileSystemSyncAccessHandle} */
  accessHandle = null;
  // The following properties are for main database files.
  /** @type {boolean} */
  isLockBusy = false;
  /** @type {boolean} */
  isFileLocked = false;
  /** @type {boolean} */
  isRequestInProgress = false;
  /** @type {function} */
  handleLockReleaser = null;
  /** @type {BroadcastChannel} */
  handleRequestChannel;
  /** @type {boolean} */
  isHandleRequested = false;
  constructor(fileHandle) {
    this.fileHandle = fileHandle;
  }
};
var OPFSCoopSyncVFS = class _OPFSCoopSyncVFS extends FacadeVFS {
  /** @type {Map<number, File>} */
  mapIdToFile = /* @__PURE__ */ new Map();
  lastError = null;
  log = null;
  //function(...args) { console.log(`[${contextName}]`, ...args) };
  /** @type {Map<string, PersistentFile>} */
  persistentFiles = /* @__PURE__ */ new Map();
  /** @type {Map<string, FileSystemSyncAccessHandle>} */
  boundAccessHandles = /* @__PURE__ */ new Map();
  /** @type {Set<FileSystemSyncAccessHandle>} */
  unboundAccessHandles = /* @__PURE__ */ new Set();
  /** @type {Set<string>} */
  accessiblePaths = /* @__PURE__ */ new Set();
  releaser = null;
  static async create(name, module) {
    const vfs = new _OPFSCoopSyncVFS(name, module);
    await Promise.all([
      vfs.isReady(),
      vfs.#initialize(DEFAULT_TEMPORARY_FILES)
    ]);
    return vfs;
  }
  constructor(name, module) {
    super(name, module);
  }
  async #initialize(nTemporaryFiles) {
    const root = await navigator.storage.getDirectory();
    for await (const entry of root.values()) {
      if (entry.kind === "directory" && entry.name.startsWith(".ahp-")) {
        await navigator.locks.request(entry.name, { ifAvailable: true }, async (lock) => {
          if (lock) {
            this.log?.(`Deleting temporary directory ${entry.name}`);
            await root.removeEntry(entry.name, { recursive: true });
          } else {
            this.log?.(`Temporary directory ${entry.name} is in use`);
          }
        });
      }
    }
    const tmpDirName = `.ahp-${Math.random().toString(36).slice(2)}`;
    this.releaser = await new Promise((resolve) => {
      navigator.locks.request(tmpDirName, () => {
        return new Promise((release) => {
          resolve(release);
        });
      });
    });
    finalizationRegistry.register(this, this.releaser);
    const tmpDir = await root.getDirectoryHandle(tmpDirName, { create: true });
    for (let i = 0; i < nTemporaryFiles; i++) {
      const tmpFile = await tmpDir.getFileHandle(`${i}.tmp`, { create: true });
      const tmpAccessHandle = await tmpFile.createSyncAccessHandle();
      this.unboundAccessHandles.add(tmpAccessHandle);
    }
  }
  /**
   * @param {string?} zName 
   * @param {number} fileId 
   * @param {number} flags 
   * @param {DataView} pOutFlags 
   * @returns {number}
   */
  jOpen(zName, fileId, flags, pOutFlags) {
    try {
      const url = new URL(zName || Math.random().toString(36).slice(2), "file://");
      const path = url.pathname;
      if (flags & SQLITE_OPEN_MAIN_DB) {
        const persistentFile = this.persistentFiles.get(path);
        if (persistentFile?.isRequestInProgress) {
          return SQLITE_BUSY;
        } else if (!persistentFile) {
          this.log?.(`creating persistent file for ${path}`);
          const create = !!(flags & SQLITE_OPEN_CREATE);
          this._module.retryOps.push((async () => {
            try {
              let dirHandle = await navigator.storage.getDirectory();
              const directories = path.split("/").filter((d) => d);
              const filename = directories.pop();
              for (const directory of directories) {
                dirHandle = await dirHandle.getDirectoryHandle(directory, { create });
              }
              for (const suffix of DB_RELATED_FILE_SUFFIXES) {
                const fileHandle = await dirHandle.getFileHandle(filename + suffix, { create });
                await this.#createPersistentFile(fileHandle);
              }
              const file2 = new File(path, flags);
              file2.persistentFile = this.persistentFiles.get(path);
              await this.#requestAccessHandle(file2);
            } catch (e) {
              const persistentFile2 = new PersistentFile(null);
              this.persistentFiles.set(path, persistentFile2);
              console.error(e);
            }
          })());
          return SQLITE_BUSY;
        } else if (!persistentFile.fileHandle) {
          this.persistentFiles.delete(path);
          return SQLITE_CANTOPEN;
        } else if (!persistentFile.accessHandle) {
          this._module.retryOps.push((async () => {
            const file2 = new File(path, flags);
            file2.persistentFile = this.persistentFiles.get(path);
            await this.#requestAccessHandle(file2);
          })());
          return SQLITE_BUSY;
        }
      }
      if (!this.accessiblePaths.has(path) && !(flags & SQLITE_OPEN_CREATE)) {
        throw new Error(`File ${path} not found`);
      }
      const file = new File(path, flags);
      this.mapIdToFile.set(fileId, file);
      if (this.persistentFiles.has(path)) {
        file.persistentFile = this.persistentFiles.get(path);
      } else if (this.boundAccessHandles.has(path)) {
        file.accessHandle = this.boundAccessHandles.get(path);
      } else if (this.unboundAccessHandles.size) {
        file.accessHandle = this.unboundAccessHandles.values().next().value;
        file.accessHandle.truncate(0);
        this.unboundAccessHandles.delete(file.accessHandle);
        this.boundAccessHandles.set(path, file.accessHandle);
      }
      this.accessiblePaths.add(path);
      pOutFlags.setInt32(0, flags, true);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_CANTOPEN;
    }
  }
  /**
   * @param {string} zName 
   * @param {number} syncDir 
   * @returns {number}
   */
  jDelete(zName, syncDir) {
    try {
      const url = new URL(zName, "file://");
      const path = url.pathname;
      if (this.persistentFiles.has(path)) {
        const persistentFile = this.persistentFiles.get(path);
        persistentFile.accessHandle.truncate(0);
      } else {
        this.boundAccessHandles.get(path)?.truncate(0);
      }
      this.accessiblePaths.delete(path);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_DELETE;
    }
  }
  /**
   * @param {string} zName 
   * @param {number} flags 
   * @param {DataView} pResOut 
   * @returns {number}
   */
  jAccess(zName, flags, pResOut) {
    try {
      const url = new URL(zName, "file://");
      const path = url.pathname;
      pResOut.setInt32(0, this.accessiblePaths.has(path) ? 1 : 0, true);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_ACCESS;
    }
  }
  /**
   * @param {number} fileId 
   * @returns {number}
   */
  jClose(fileId) {
    try {
      const file = this.mapIdToFile.get(fileId);
      this.mapIdToFile.delete(fileId);
      if (file?.flags & SQLITE_OPEN_MAIN_DB) {
        if (file.persistentFile?.handleLockReleaser) {
          this.#releaseAccessHandle(file);
        }
      } else if (file?.flags & SQLITE_OPEN_DELETEONCLOSE) {
        file.accessHandle.truncate(0);
        this.accessiblePaths.delete(file.path);
        if (!this.persistentFiles.has(file.path)) {
          this.boundAccessHandles.delete(file.path);
          this.unboundAccessHandles.add(file.accessHandle);
        }
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_CLOSE;
    }
  }
  /**
   * @param {number} fileId 
   * @param {Uint8Array} pData 
   * @param {number} iOffset
   * @returns {number}
   */
  jRead(fileId, pData, iOffset) {
    try {
      const file = this.mapIdToFile.get(fileId);
      const accessHandle = file.accessHandle || file.persistentFile.accessHandle;
      const bytesRead = accessHandle.read(pData.subarray(), { at: iOffset });
      if (file.flags & SQLITE_OPEN_MAIN_DB && !file.persistentFile.isFileLocked) {
        this.#releaseAccessHandle(file);
      }
      if (bytesRead < pData.byteLength) {
        pData.fill(0, bytesRead);
        return SQLITE_IOERR_SHORT_READ;
      }
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_READ;
    }
  }
  /**
   * @param {number} fileId 
   * @param {Uint8Array} pData 
   * @param {number} iOffset
   * @returns {number}
   */
  jWrite(fileId, pData, iOffset) {
    try {
      const file = this.mapIdToFile.get(fileId);
      const accessHandle = file.accessHandle || file.persistentFile.accessHandle;
      const nBytes = accessHandle.write(pData.subarray(), { at: iOffset });
      if (nBytes !== pData.byteLength) throw new Error("short write");
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_WRITE;
    }
  }
  /**
   * @param {number} fileId 
   * @param {number} iSize 
   * @returns {number}
   */
  jTruncate(fileId, iSize) {
    try {
      const file = this.mapIdToFile.get(fileId);
      const accessHandle = file.accessHandle || file.persistentFile.accessHandle;
      accessHandle.truncate(iSize);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_TRUNCATE;
    }
  }
  /**
   * @param {number} fileId 
   * @param {number} flags 
   * @returns {number}
   */
  jSync(fileId, flags) {
    try {
      const file = this.mapIdToFile.get(fileId);
      const accessHandle = file.accessHandle || file.persistentFile.accessHandle;
      accessHandle.flush();
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_FSYNC;
    }
  }
  /**
   * @param {number} fileId 
   * @param {DataView} pSize64 
   * @returns {number}
   */
  jFileSize(fileId, pSize64) {
    try {
      const file = this.mapIdToFile.get(fileId);
      const accessHandle = file.accessHandle || file.persistentFile.accessHandle;
      const size = accessHandle.getSize();
      pSize64.setBigInt64(0, BigInt(size), true);
      return SQLITE_OK;
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR_FSTAT;
    }
  }
  /**
   * @param {number} fileId 
   * @param {number} lockType 
   * @returns {number}
   */
  jLock(fileId, lockType) {
    const file = this.mapIdToFile.get(fileId);
    if (file.persistentFile.isRequestInProgress) {
      file.persistentFile.isLockBusy = true;
      return SQLITE_BUSY;
    }
    file.persistentFile.isFileLocked = true;
    if (!file.persistentFile.handleLockReleaser) {
      file.persistentFile.handleRequestChannel.onmessage = () => {
        this.log?.(`received notification for ${file.path}`);
        if (file.persistentFile.isFileLocked) {
          file.persistentFile.isHandleRequested = true;
        } else {
          this.#releaseAccessHandle(file);
        }
        file.persistentFile.handleRequestChannel.onmessage = null;
      };
      this.#requestAccessHandle(file);
      this.log?.("returning SQLITE_BUSY");
      file.persistentFile.isLockBusy = true;
      return SQLITE_BUSY;
    }
    file.persistentFile.isLockBusy = false;
    return SQLITE_OK;
  }
  /**
   * @param {number} fileId 
   * @param {number} lockType 
   * @returns {number}
   */
  jUnlock(fileId, lockType) {
    const file = this.mapIdToFile.get(fileId);
    if (lockType === SQLITE_LOCK_NONE) {
      if (!file.persistentFile.isLockBusy) {
        if (file.persistentFile.isHandleRequested) {
          this.#releaseAccessHandle(file);
          file.persistentFile.isHandleRequested = false;
        }
        file.persistentFile.isFileLocked = false;
      }
    }
    return SQLITE_OK;
  }
  /**
   * @param {number} fileId
   * @param {number} op
   * @param {DataView} pArg
   * @returns {number|Promise<number>}
   */
  jFileControl(fileId, op, pArg) {
    try {
      const file = this.mapIdToFile.get(fileId);
      switch (op) {
        case SQLITE_FCNTL_PRAGMA:
          const key = extractString(pArg, 4);
          const value = extractString(pArg, 8);
          this.log?.("xFileControl", file.path, "PRAGMA", key, value);
          switch (key.toLowerCase()) {
            case "journal_mode":
              if (value && !["off", "memory", "delete", "wal"].includes(value.toLowerCase())) {
                throw new Error('journal_mode must be "off", "memory", "delete", or "wal"');
              }
              break;
          }
          break;
      }
    } catch (e) {
      this.lastError = e;
      return SQLITE_IOERR;
    }
    return SQLITE_NOTFOUND;
  }
  /**
   * @param {Uint8Array} zBuf 
   * @returns 
   */
  jGetLastError(zBuf) {
    if (this.lastError) {
      console.error(this.lastError);
      const outputArray = zBuf.subarray(0, zBuf.byteLength - 1);
      const { written } = new TextEncoder().encodeInto(this.lastError.message, outputArray);
      zBuf[written] = 0;
    }
    return SQLITE_OK;
  }
  /**
   * @param {FileSystemFileHandle} fileHandle 
   * @returns {Promise<PersistentFile>}
   */
  async #createPersistentFile(fileHandle) {
    const persistentFile = new PersistentFile(fileHandle);
    const root = await navigator.storage.getDirectory();
    const relativePath = await root.resolve(fileHandle);
    const path = `/${relativePath.join("/")}`;
    persistentFile.handleRequestChannel = new BroadcastChannel(`ahp:${path}`);
    this.persistentFiles.set(path, persistentFile);
    const f = await fileHandle.getFile();
    if (f.size) {
      this.accessiblePaths.add(path);
    }
    return persistentFile;
  }
  /**
   * @param {File} file 
   */
  #requestAccessHandle(file) {
    console.assert(!file.persistentFile.handleLockReleaser);
    if (!file.persistentFile.isRequestInProgress) {
      file.persistentFile.isRequestInProgress = true;
      this._module.retryOps.push((async () => {
        file.persistentFile.handleLockReleaser = await this.#acquireLock(file.persistentFile);
        this.log?.(`creating access handles for ${file.path}`);
        await Promise.all(DB_RELATED_FILE_SUFFIXES.map(async (suffix) => {
          const persistentFile = this.persistentFiles.get(file.path + suffix);
          if (persistentFile) {
            persistentFile.accessHandle = await persistentFile.fileHandle.createSyncAccessHandle();
          }
        }));
        file.persistentFile.isRequestInProgress = false;
      })());
      return this._module.retryOps.at(-1);
    }
    return Promise.resolve();
  }
  /**
   * @param {File} file 
   */
  async #releaseAccessHandle(file) {
    DB_RELATED_FILE_SUFFIXES.forEach(async (suffix) => {
      const persistentFile = this.persistentFiles.get(file.path + suffix);
      if (persistentFile) {
        persistentFile.accessHandle?.close();
        persistentFile.accessHandle = null;
      }
    });
    this.log?.(`access handles closed for ${file.path}`);
    file.persistentFile.handleLockReleaser?.();
    file.persistentFile.handleLockReleaser = null;
    this.log?.(`lock released for ${file.path}`);
  }
  /**
   * @param {PersistentFile} persistentFile 
   * @returns  {Promise<function>} lock releaser
   */
  #acquireLock(persistentFile) {
    return new Promise((resolve) => {
      const lockName = persistentFile.handleRequestChannel.name;
      const notify = () => {
        this.log?.(`notifying for ${lockName}`);
        persistentFile.handleRequestChannel.postMessage(null);
      };
      const notifyId = setInterval(notify, LOCK_NOTIFY_INTERVAL);
      setTimeout(notify);
      this.log?.(`lock requested: ${lockName}`);
      navigator.locks.request(lockName, (lock) => {
        this.log?.(`lock acquired: ${lockName}`, lock);
        clearInterval(notifyId);
        return new Promise(resolve);
      });
    });
  }
};
function extractString(dataView, offset) {
  const p = dataView.getUint32(offset, true);
  if (p) {
    const chars = new Uint8Array(dataView.buffer, p);
    return new TextDecoder().decode(chars.subarray(0, chars.indexOf(0)));
  }
  return null;
}

// src/vfs/opfs.ts
async function useOpfsStorage(path, options = {}) {
  const { url, ...rest } = options;
  const sqliteModule = await wa_sqlite_default(
    options.url ? { locateFile: () => options.url } : void 0
  );
  return {
    path,
    sqliteModule,
    vfsFn: OPFSCoopSyncVFS.create,
    ...rest
  };
}

export { OPFSCoopSyncVFS, useOpfsStorage };
